import os
import sys
import unittest
from io import StringIO
from unittest.mock import patch, Mock

import click as click

from safety import util
from safety.util import read_requirements, get_processed_options, SafetyPolicyFile


class ReadRequirementsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dirname = os.path.dirname(__file__)

    def test_unpinned_vcs_requirement(self):
        """
        https://github.com/pyupio/safety/issues/72
        """
        # this shouldn't raise an error
        content = StringIO("-e git+https://github.com/jdunck/python-unicodecsv#egg=unicodecsv")
        result = list(read_requirements(content))
        self.assertEqual(len(result), 0)

    def test_recursive_requirement(self):
        """
        https://github.com/pyupio/safety/issues/132
        """
        # this should find 2 bad packages
        dirname = os.path.dirname(__file__)
        test_filename = os.path.join(dirname, "reqs_1.txt")
        with open(test_filename) as fh:
            result = list(read_requirements(fh, resolve=True))
        self.assertEqual(len(result), 2)

    def test_recursive_requirement_pinned_after_unpinned(self):
        # this should find 3 packages, unpinned are ignored
        dirname = os.path.dirname(__file__)
        test_filename = os.path.join(dirname, "reqs_pinned_and_unpinned.txt")
        with open(test_filename) as fh:
            result = list(read_requirements(fh, resolve=True))
        self.assertEqual(len(result), 3)

    @patch("safety.util.get_flags_from_context")
    @patch.object(sys, 'argv', ['safety/__main__.py', 'check', '--key=my-key', '-i', '3232', '-i', '3231', '--ignore',
                                '1212', '--json'])
    def test_log_used_options_with_argv(self, get_flags_from_context):
        get_flags_from_context.return_value = {'--key': 'key', '--db': 'db', '--json': 'json', '--no-json': 'json',
                                               '--full-report': 'full_report', '--short-report': 'full_report',
                                               '--bare': 'bare', '--not-bare': 'bare', '--cache': 'cache',
                                               '--no-cache': 'cache', '--stdin': 'stdin', '--no-stdin': 'stdin',
                                               '--file': 'files', '-r': 'files', '--ignore': 'ignore', '-i': 'ignore',
                                               '--output': 'output', '-o': 'output', '--proxy-host': 'proxyhost',
                                               '-ph': 'proxyhost', '--proxy-port': 'proxyport', '-pp': 'proxyport',
                                               '--proxy-protocol': 'proxyprotocol', '-pr': 'proxyprotocol'}
        used_options = util.get_used_options()

        self.assertEqual(used_options, {'ignore': {'-i': 2, '--ignore': 1}, 'json': {'--json': 1}, 'key': {'--key': 1}})

    @patch.object(click, 'get_current_context',
                  Mock(get_parameter_source=Mock(return_value=click.core.ParameterSource.DEFAULT)))
    def test_cli_ignore_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_with_ignores.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        cli_ignores = {'1234': {'reason': '', 'expires': None}}
        ignore, ignore_severity_rules, exit_code = get_processed_options(policy_file=policy_file, ignore=cli_ignores,
                                                                         ignore_severity_rules=None, exit_code=True)

        self.assertEqual(ignore, cli_ignores)

    @patch.object(click, 'get_current_context',
                  Mock(get_parameter_source=Mock(return_value=click.core.ParameterSource.COMMANDLINE)))
    def test_cli_continue_on_error_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        ignore, ignore_severity_rules, exit_code = get_processed_options(policy_file=policy_file, ignore={},
                                                                         ignore_severity_rules=None, exit_code=True)

        self.assertEqual(exit_code, True)

    @patch.object(click, 'get_current_context',
                  Mock(get_parameter_source=Mock(return_value=click.core.ParameterSource.COMMANDLINE)))
    def test_cli_exit_code_partial_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        # Cli only passes the exit_code argument by commandline
        ignore, ignore_severity_rules, exit_code = get_processed_options(policy_file=policy_file, ignore={},
                                                                         ignore_severity_rules=None, exit_code=True)

        security_pf = policy_file.get('security', {})
        severity_rules = {'ignore-cvss-severity-below': security_pf.get('ignore-cvss-severity-below', 0.0),
                          'ignore-cvss-unknown-severity': security_pf.get('ignore-cvss-unknown-severity', False)}

        self.assertEqual(ignore, security_pf.get('ignore-vulnerabilities', None))
        self.assertEqual(ignore_severity_rules, severity_rules)
        self.assertEqual(exit_code, True)

    @patch.object(click, 'get_current_context',
                  Mock(get_parameter_source=Mock(return_value=click.core.ParameterSource.DEFAULT)))
    def test_cli_ignore_partial_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        # Cli only passes the ignores argument by commandline
        cli_ignores = {'1234': {'reason': '', 'expires': None}}
        ignore, ignore_severity_rules, exit_code = get_processed_options(policy_file=policy_file, ignore=cli_ignores,
                                                                         ignore_severity_rules=None, exit_code=True)

        security_pf = policy_file.get('security', {})
        severity_rules = {'ignore-cvss-severity-below': security_pf.get('ignore-cvss-severity-below', 0.0),
                          'ignore-cvss-unknown-severity': security_pf.get('ignore-cvss-unknown-severity', False)}

        self.assertEqual(ignore, cli_ignores)
        self.assertEqual(ignore_severity_rules, severity_rules)
        self.assertIsNotNone(security_pf.get('continue-on-vulnerability-error', None),
                             msg='This test requires a yml with a continue-on-vulnerability-error value')
        EXPECTED = not security_pf.get('continue-on-vulnerability-error')
        self.assertEqual(exit_code, EXPECTED)






import json
import os
import tempfile
import unittest
from datetime import datetime
from unittest.mock import patch, Mock

import click
from click.testing import CliRunner

from safety import cli
from safety.models import Vulnerability, CVE, Severity
from safety.util import Package, SafetyContext


def get_vulnerability(vuln_kwargs=None, cve_kwargs=None, pkg_kwargs=None):
    vuln_kwargs = {} if vuln_kwargs is None else vuln_kwargs
    cve_kwargs = {} if cve_kwargs is None else cve_kwargs
    pkg_kwargs = {} if pkg_kwargs is None else pkg_kwargs

    p_kwargs = {'name': 'django', 'version': '2.2', 'found': '/site-packages/django', 'insecure_versions': [],
                'secure_versions': ['2.2'], 'latest_version_without_known_vulnerabilities': '2.2',
                'latest_version': '2.2', 'more_info_url': 'https://pyup.io/package/foo'}
    p_kwargs.update(pkg_kwargs)

    c_kwargs = {'name': 'bla', 'cvssv2': None, 'cvssv3': None}
    c_kwargs.update(cve_kwargs)

    cve = CVE(**c_kwargs)
    severity = None
    if cve and cve.cvssv2 or cve.cvssv3:
        severity = Severity(source=cve.name, cvssv2=cve.cvssv2, cvssv3=cve.cvssv3)
    pkg = Package(**p_kwargs)

    v_kwargs = {'package_name': pkg.name, 'pkg': pkg, 'ignored': False, 'ignored_reason': '', 'ignored_expires': '',
                'vulnerable_spec': ">0",
                'all_vulnerable_specs': ['2.2'],
                'analyzed_version': pkg.version,
                'advisory': '',
                'vulnerability_id': 'PYUP-1234',
                'is_transitive': False,
                'published_date': '',
                'fixed_versions': [],
                'closest_versions_without_known_vulnerabilities': '',
                'resources': ["pyup.io/vuln-id"],
                'CVE': cve,
                'severity': severity,
                'affected_versions': [],
                'more_info_url': 'https://pyup.io/PVE/2323'}

    v_kwargs.update(vuln_kwargs)

    return Vulnerability(**v_kwargs)


class TestSafetyCLI(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.runner = CliRunner(mix_stderr=False)
        self.output_options = ['screen', 'text', 'json', 'bare']
        self.dirname = os.path.dirname(__file__)

    def test_command_line_interface(self):
        runner = CliRunner()
        result = runner.invoke(cli.cli)
        assert result.exit_code == 0
        assert 'Usage:' in result.output

        help_result = runner.invoke(cli.cli, ['--help'])
        assert help_result.exit_code == 0
        assert '--help' in help_result.output

    @patch("safety.safety.check")
    def test_check_vulnerabilities_found_default(self, check_func):
        check_func.return_value = [get_vulnerability()], None
        EXPECTED_EXIT_CODE_VULNS_FOUND = 64
        result = self.runner.invoke(cli.cli, ['check'])
        self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_VULNS_FOUND)

    @patch("safety.safety.check")
    def test_check_vulnerabilities_not_found_default(self, check_func):
        check_func.return_value = [], None
        EXPECTED_EXIT_CODE_VULNS_NOT_FOUND = 0
        result = self.runner.invoke(cli.cli, ['check'])
        self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_VULNS_NOT_FOUND)

    @patch("safety.safety.check")
    def test_check_vulnerabilities_found_with_outputs(self, check_func):
        check_func.return_value = [get_vulnerability()], None
        EXPECTED_EXIT_CODE_VULNS_FOUND = 64

        for output in self.output_options:
            result = self.runner.invoke(cli.cli, ['check', '--output', output])
            self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_VULNS_FOUND)

    @patch("safety.safety.check")
    def test_check_vulnerabilities_not_found_with_outputs(self, check_func):
        check_func.return_value = [], None
        EXPECTED_EXIT_CODE_VULNS_NOT_FOUND = 0

        for output in self.output_options:
            result = self.runner.invoke(cli.cli, ['check', '--output', output])
            self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_VULNS_NOT_FOUND)

    @patch("safety.safety.check")
    def test_check_continue_on_error(self, check_func):
        EXPECTED_EXIT_CODE_CONTINUE_ON_ERROR = 0

        # Cover no vulns found and vulns found
        for vulns in [[get_vulnerability()], []]:
            check_func.return_value = vulns, None

            result = self.runner.invoke(cli.cli, ['check', '--continue-on-error'])
            self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_CONTINUE_ON_ERROR)

            for output in self.output_options:
                result = self.runner.invoke(cli.cli, ['check', '--output', output, '--continue-on-error'])
                self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_CONTINUE_ON_ERROR)

    @patch("safety.safety.get_announcements")
    def test_announcements_if_is_not_tty(self, get_announcements_func):
        announcement = {'type': 'error', 'message': 'Please upgrade now'}
        get_announcements_func.return_value = [announcement]
        message = f"* {announcement.get('message')}"
        result = self.runner.invoke(cli.cli, ['check'])
        self.assertTrue('ANNOUNCEMENTS' in result.stderr)
        self.assertTrue(message in result.stderr)

    @patch("safety.safety.get_announcements")
    def test_review_pass(self, mocked_announcements):
        mocked_announcements.return_value = []
        runner = CliRunner()
        dirname = os.path.dirname(__file__)
        path_to_report = os.path.join(dirname, "test_db", "report.json")
        result = runner.invoke(cli.cli, ['review', '--output', 'bare', '--file', path_to_report])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, u'insecure-package\n')

    @patch("safety.util.SafetyContext")
    @patch("safety.safety.check")
    @patch("safety.cli.get_packages")
    def test_chained_review_pass(self, get_packages, check_func, ctx):
        expires = datetime.strptime('2022-10-21', '%Y-%m-%d')
        vulns = [get_vulnerability(), get_vulnerability(vuln_kwargs={'vulnerability_id': '25853', 'ignored': True,
                                                                     'ignored_reason': 'A basic reason',
                                                                     'ignored_expires': expires})]
        packages = [pkg for pkg in {vuln.pkg.name: vuln.pkg for vuln in vulns}.values()]
        get_packages.return_value = packages
        provided_context = SafetyContext()
        provided_context.command = 'check'
        provided_context.packages = packages
        ctx.return_value = provided_context
        check_func.return_value = vulns, None

        with tempfile.TemporaryDirectory() as tempdir:
            for output in self.output_options:
                path_to_report = os.path.join(tempdir, f'report_{output}.json')

                pre_result = self.runner.invoke(cli.cli, [
                    'check', '--key', 'foo', '-o', output,
                    '--save-json', path_to_report])

                self.assertEqual(pre_result.exit_code, 64)

            for output in self.output_options:
                filename = f'report_{output}.json'
                path_to_report = os.path.join(tempdir, filename)
                result = self.runner.invoke(cli.cli, ['review', '--output', output, '--file', path_to_report])
                self.assertEqual(result.exit_code, 0, f'Unable to load the previous saved report: {filename}')

    @patch("safety.safety.session")
    def test_license_with_file(self, requests_session):
        licenses_db = {
            "licenses": {
                "BSD-3-Clause": 2
            },
            "packages": {
                "django": [
                    {
                        "start_version": "0.0",
                        "license_id": 2
                    }
                ]
            }
        }

        mock = Mock()
        mock.json.return_value = licenses_db
        mock.status_code = 200
        requests_session.get.return_value = mock

        dirname = os.path.dirname(__file__)
        test_filename = os.path.join(dirname, "reqs_4.txt")
        result = self.runner.invoke(cli.cli, ['license', '--key', 'foo', '--file', test_filename])
        # TODO: Add test for the screen formatter, this only test that the command doesn't crash
        self.assertEqual(result.exit_code, 0)

    def test_validate_with_unsupported_argument(self):
        result = self.runner.invoke(cli.cli, ['validate', 'safety_ci'])
        msg = 'This Safety version only supports "policy_file" validation. "safety_ci" is not supported.\n'
        self.assertEqual(click.unstyle(result.stderr), msg)
        self.assertEqual(result.exit_code, 1)

    def test_validate_with_wrong_path(self):
        result = self.runner.invoke(cli.cli, ['validate', 'policy_file', '--path', 'imaginary/path'])
        msg = 'The path "imaginary/path" does not exist.\n'
        self.assertEqual(click.unstyle(result.stderr), msg)
        self.assertEqual(result.exit_code, 1)

    def test_validate_with_basic_policy_file(self):
        dirname = os.path.dirname(__file__)
        path = os.path.join(dirname, "test_policy_file", "default_policy_file.yml")
        result = self.runner.invoke(cli.cli, ['validate', 'policy_file', '--path', path])
        cleaned_stdout = click.unstyle(result.stdout)
        msg = 'The Safety policy file was successfully parsed with the following values:\n'
        parsed = json.dumps(
            {
                "security": {
                    "ignore-cvss-severity-below": 0,
                    "ignore-cvss-unknown-severity": False,
                    "ignore-vulnerabilities": {
                        "25853": {
                            "reason": "we don't use the vulnerable function",
                            "expires": "2022-10-21 00:00:00"
                        }
                    },
                    "continue-on-vulnerability-error": False
                },
                "filename": path
            },
            indent=4
        ) + '\n'

        self.assertEqual(msg + parsed, cleaned_stdout)
        self.assertEqual(result.exit_code, 0)

    def test_validate_with_policy_file_using_invalid_keyword(self):
        dirname = os.path.dirname(__file__)
        filename = 'default_policy_file_using_invalid_keyword.yml'
        path = os.path.join(dirname, "test_policy_file", filename)
        result = self.runner.invoke(cli.cli, ['validate', 'policy_file', '--path', path])
        cleaned_stdout = click.unstyle(result.stderr)
        msg_hint = 'HINT: "security" -> "transitive" is not a valid keyword. Valid keywords in this level are: ' \
                   'ignore-cvss-severity-below, ignore-cvss-unknown-severity, ignore-vulnerabilities, ' \
                   'continue-on-vulnerability-error\n'
        msg = f'Unable to load the Safety Policy file "{path}".\n{msg_hint}'

        self.assertEqual(msg, cleaned_stdout)
        self.assertEqual(result.exit_code, 1)

    def test_validate_with_policy_file_using_invalid_typo_keyword(self):
        dirname = os.path.dirname(__file__)
        filename = 'default_policy_file_using_invalid_typo_keyword.yml'
        path = os.path.join(dirname, "test_policy_file", filename)
        result = self.runner.invoke(cli.cli, ['validate', 'policy_file', '--path', path])
        cleaned_stdout = click.unstyle(result.stderr)
        msg_hint = 'HINT: "security" -> "ignore-vunerabilities" is not a valid keyword. Maybe you meant: ' \
                   'ignore-vulnerabilities\n'
        msg = f'Unable to load the Safety Policy file "{path}".\n{msg_hint}'

        self.assertEqual(msg, cleaned_stdout)
        self.assertEqual(result.exit_code, 1)

    def test_generate_pass(self):
        with tempfile.TemporaryDirectory() as tempdir:
            result = self.runner.invoke(cli.cli, ['generate', 'policy_file', '--path', tempdir])
            cleaned_stdout = click.unstyle(result.stdout)
            msg = f'A default Safety policy file has been generated! Review the file contents in the path {tempdir} ' \
                  f'in the file: .safety-policy.yml\n'
            self.assertEqual(msg, cleaned_stdout)

    def test_generate_with_unsupported_argument(self):
        result = self.runner.invoke(cli.cli, ['generate', 'safety_ci'])
        msg = 'This Safety version only supports "policy_file" generation. "safety_ci" is not supported.\n'
        self.assertEqual(click.unstyle(result.stderr), msg)
        self.assertEqual(result.exit_code, 1)

    def test_generate_with_wrong_path(self):
        result = self.runner.invoke(cli.cli, ['generate', 'policy_file', '--path', 'imaginary/path'])
        msg = 'The path "imaginary/path" does not exist.\n'
        self.assertEqual(click.unstyle(result.stderr), msg)
        self.assertEqual(result.exit_code, 1)









from safety.models import Package, Vulnerability, CVE, Severity

VALID_REPORT = {'report_meta': {'scan_target': 'environment', 'scanned': ['/usr/local/lib/python3.9/site-packages'],
                                'api_key': True, 'packages_found': 50, 'local_database_path_used': None,
                                'timestamp': '2022-04-30 08:16:30',
                                'safety_version': '2.0b1'},
                'scanned_packages': {'safety': {'name': 'safety', 'version': '2.0b1'},
                                     'click': {'name': 'click', 'version': '8.0.4'},
                                     'insecure-package': {'name': 'insecure-package', 'version': '0.1.0'}},
                'affected_packages': {'insecure-package': {'name': 'insecure-package', 'version': '0.1.0',
                                                           'found': '/usr/local/lib/python3.9/site-packages',
                                                           'insecure_versions': ['0.1.0'], 'secure_versions': [],
                                                           'latest_version_without_known_vulnerabilities': '',
                                                           'latest_version': '0.1.0',
                                                           'more_info_url': 'https://pyup.io/packages/pypi/insecure-package/'}},
                'announcements': [], 'vulnerabilities': [
        {'vulnerability_id': '25853', 'package_name': 'insecure-package', 'ignored': False, 'ignored_reason': None,
         'ignored_expires': None, 'vulnerable_spec': '<0.2.0',
         'all_vulnerable_specs': ['<0.2.0'], 'analyzed_version': '0.1.0',
         'advisory': 'This is an insecure package with lots of exploitable security vulnerabilities.',
         'is_transitive': False, 'published_date': '2021-Apr-14',
         'fixed_versions': [''], 'closest_versions_without_known_vulnerabilities': [],
         'resources': ['https://pypi.org/project/insecure-package'],
         'CVE': None,
         'severity': None, 'affected_versions': ['0.1.0'],
         'more_info_url': 'https://pyup.io/vulnerabilities/PVE-2021-25853/25853/'}],
                'ignored_vulnerabilities': [], 'remediations': {
        'insecure-package': {'current_version': '0.1.0', 'vulnerabilities_found': 1, 'recommended_version': None,
                             'other_recommended_versions': [],
                             'more_info_url': 'https://pyup.io/packages/pypi/insecure-package/'}}}

SCANNED_PACKAGES = {
    'safety': Package(name='safety', version='2.0b1'),
    'click': Package(name='click', version='8.0.4'),
    'insecure-package': Package(name='insecure-package', version='0.1.0'),
}

VULNS = [Vulnerability(vulnerability_id='25853', package_name='insecure-package',
                       pkg=Package(name='insecure-package', version='0.1.0',
                                   found=None,
                                   insecure_versions=None, secure_versions=None,
                                   latest_version_without_known_vulnerabilities=None,
                                   latest_version=None,
                                   more_info_url=None),
                       ignored=False, ignored_reason=None, ignored_expires=None, vulnerable_spec='<0.2.0',
                       all_vulnerable_specs=['<0.2.0'],
                       analyzed_version='0.1.0',
                       advisory='This is an insecure package with lots of exploitable security vulnerabilities.',
                       is_transitive=False, published_date='2021-Apr-14', fixed_versions=[''],
                       closest_versions_without_known_vulnerabilities=[],
                       resources=['https://pypi.org/project/insecure-package'],
                       CVE=None, severity=None, affected_versions=['0.1.0'],
                       more_info_url='https://pyup.io/vulnerabilities/PVE-2021-25853/25853/')]

REMEDIATIONS = {'insecure-package': {'vulns_found': 1, 'version': '0.1.0', 'secure_versions': [],
                                     'closest_secure_version': {'major': None, 'minor': None},
                                     'more_info_url': 'https://pyup.io/packages/pypi/insecure-package/'}}
